@@ -147,8 +147,7 @@ void amqp_envelope_dtor(void *object TSRMLS_DC)
 	amqp_envelope_object *envelope = (amqp_envelope_object*)object;
 
 	if (envelope->headers) {
-		zval_dtor(envelope->headers);
-		efree(envelope->headers);
+		zval_ptr_dtor(&envelope->headers);
 	}
 
 	if (envelope->body) {
@@ -435,10 +434,6 @@ PHP_METHOD(amqp_envelope_class, getExpiration)
 	/* Get the envelope object out of the store */
 	envelope = (amqp_envelope_object *)zend_object_store_get_object(id TSRMLS_CC);
 
-	if (envelope->expiration == 0) {
-		RETURN_FALSE;
-	}
-
 	RETURN_STRING(envelope->expiration, 1);
 }
 /* }}} */
@@ -584,11 +579,8 @@ PHP_METHOD(amqp_envelope_class, getHeaders)
 	/* Get the envelope object out of the store */
 	envelope = (amqp_envelope_object *)zend_object_store_get_object(id TSRMLS_CC);
 
-	*return_value = *envelope->headers;
-	zval_copy_ctor(return_value);
-
-	/* Increment the ref count */
-	Z_ADDREF_P(envelope->headers);
+	zval_dtor(return_value);
+	MAKE_COPY_ZVAL(&envelope->headers, return_value);
 }
 /* }}} */
 
