@@ -65,7 +65,7 @@ HashTable *amqp_exchange_object_get_debug_info(zval *object, int *is_temp TSRMLS
 
 	/* Keep the first number matching the number of entries in this table*/
 	ALLOC_HASHTABLE(debug_info);
-	ZEND_INIT_SYMTABLE_EX(debug_info, 5 + 1, 0);
+	ZEND_INIT_SYMTABLE_EX(debug_info, 6 + 1, 0);
 
 	/* Start adding values */
 	MAKE_STD_ZVAL(value);
@@ -77,17 +77,22 @@ HashTable *amqp_exchange_object_get_debug_info(zval *object, int *is_temp TSRMLS
 	zend_hash_add(debug_info, "type", sizeof("type"), &value, sizeof(zval *), NULL);
 
 	MAKE_STD_ZVAL(value);
-	ZVAL_LONG(value, exchange->passive);
+	ZVAL_BOOL(value, IS_PASSIVE(exchange->flags));
 	zend_hash_add(debug_info, "passive", sizeof("passive"), &value, sizeof(zval *), NULL);
 
 	MAKE_STD_ZVAL(value);
-	ZVAL_LONG(value, exchange->durable);
+	ZVAL_BOOL(value, IS_DURABLE(exchange->flags));
 	zend_hash_add(debug_info, "durable", sizeof("durable"), &value, sizeof(zval *), NULL);
-	
+
 	MAKE_STD_ZVAL(value);
-	ZVAL_LONG(value, exchange->auto_delete);
+	ZVAL_BOOL(value, IS_AUTODELETE(exchange->flags));
 	zend_hash_add(debug_info, "auto_delete", sizeof("auto_delete"), &value, sizeof(zval *), NULL);
 
+	MAKE_STD_ZVAL(value);
+	ZVAL_BOOL(value, IS_INTERNAL(exchange->flags));
+	zend_hash_add(debug_info, "internal", sizeof("internal"), &value, sizeof(zval *), NULL);
+
+	Z_ADDREF_P(exchange->arguments);
 	zend_hash_add(debug_info, "arguments", sizeof("arguments"), &exchange->arguments, sizeof(&exchange->arguments), NULL);
 
 	/* Start adding values */
@@ -145,30 +150,6 @@ zend_object_value amqp_exchange_ctor(zend_class_entry *ce TSRMLS_DC)
 	return new_value;
 }
 
-void free_field_value(struct amqp_field_value_t_ value) {
-	switch (value.kind) {
-		case AMQP_FIELD_KIND_ARRAY:
-			{
-				int i;
-				for (i=0; i<value.value.array.num_entries; ++i) {
-					free_field_value(value.value.array.entries[i]);
-				}
-				efree(value.value.array.entries);
-			}
-			break;
-		case AMQP_FIELD_KIND_TABLE:
-			{
-				int i;
-				for (i=0; i<value.value.table.num_entries; ++i) {
-					free_field_value(value.value.table.entries[i].value);
-				}
-				efree(value.value.table.entries);
-			}
-			break;
-	}
-}
-
-
 /* {{{ proto AMQPExchange::__construct(AMQPChannel channel);
 create Exchange   */
 PHP_METHOD(amqp_exchange_class, __construct)
@@ -197,11 +178,7 @@ PHP_METHOD(amqp_exchange_class, __construct)
 
 	/* Pull the channel out */
 	channel = AMQP_GET_CHANNEL(exchange);
-
 	AMQP_VERIFY_CHANNEL(channel, "Could not create exchange.");
-
-	/* We have a valid connection: */
-	exchange->is_connected = '\1';
 }
 /* }}} */
 
@@ -263,7 +240,6 @@ PHP_METHOD(amqp_exchange_class, getFlags)
 {
 	zval *id;
 	amqp_exchange_object *exchange;
-	long flagBitmask = 0;
 
 	if (zend_parse_method_parameters(ZEND_NUM_ARGS() TSRMLS_CC, getThis(), "O", &id, amqp_exchange_class_entry) == FAILURE) {
 		return;
@@ -271,12 +247,7 @@ PHP_METHOD(amqp_exchange_class, getFlags)
 
 	exchange = (amqp_exchange_object *)zend_object_store_get_object(id TSRMLS_CC);
 
-	/* Set the bitmask based on what is set in the exchange */
-	flagBitmask |= (exchange->passive ? AMQP_PASSIVE : 0);
-	flagBitmask |= (exchange->durable ? AMQP_DURABLE : 0);
-	flagBitmask |= (exchange->auto_delete ? AMQP_AUTODELETE : 0);
-
-	RETURN_LONG(flagBitmask);
+	RETURN_LONG(exchange->flags);
 }
 /* }}} */
 
@@ -297,9 +268,7 @@ PHP_METHOD(amqp_exchange_class, setFlags)
 	exchange = (amqp_exchange_object *)zend_object_store_get_object(id TSRMLS_CC);
 
 	/* Set the flags based on the bitmask we were given */
-	exchange->passive = IS_PASSIVE(flagBitmask);
-	exchange->durable = IS_DURABLE(flagBitmask);
-	exchange->auto_delete = IS_AUTODELETE(flagBitmask);
+	exchange->flags = flagBitmask ? flagBitmask & PHP_AMQP_EXCHANGE_FLAGS : flagBitmask;
 }
 /* }}} */
 
@@ -342,13 +311,12 @@ PHP_METHOD(amqp_exchange_class, setType)
 
 	/* Pull the exchange off the object store */
 	exchange = (amqp_exchange_object *)zend_object_store_get_object(id TSRMLS_CC);
-	
+
 	AMQP_SET_TYPE(exchange, type);
 }
 /* }}} */
 
 
-
 /* {{{ proto AMQPExchange::getArgument(string key)
 Get the exchange argument referenced by key */
 PHP_METHOD(amqp_exchange_class, getArgument)
@@ -370,12 +338,13 @@ PHP_METHOD(amqp_exchange_class, getArgument)
 	}
 
 	*return_value = **tmp;
-	
+
 	zval_copy_ctor(return_value);
 	INIT_PZVAL(return_value);
 }
 /* }}} */
 
+
 /* {{{ proto AMQPExchange::getArguments
 Get the exchange arguments */
 PHP_METHOD(amqp_exchange_class, getArguments)
@@ -389,11 +358,8 @@ PHP_METHOD(amqp_exchange_class, getArguments)
 
 	exchange = (amqp_exchange_object *)zend_object_store_get_object(id TSRMLS_CC);
 
-	*return_value = *exchange->arguments;
-	zval_copy_ctor(return_value);
-
-	/* Increment the ref count */
-	Z_ADDREF_P(exchange->arguments);
+	zval_dtor(return_value);
+	MAKE_COPY_ZVAL(&exchange->arguments, return_value);
 }
 /* }}} */
 
@@ -474,9 +440,8 @@ PHP_METHOD(amqp_exchange_class, declareExchange)
 	amqp_exchange_object *exchange;
 	amqp_channel_object *channel;
 	amqp_connection_object *connection;
-	amqp_table_t *arguments;
 
-	amqp_rpc_reply_t res;
+	amqp_table_t *arguments;
 
 	if (zend_parse_method_parameters(ZEND_NUM_ARGS() TSRMLS_CC, getThis(), "O", &id, amqp_exchange_class_entry) == FAILURE) {
 		return;
@@ -502,33 +467,51 @@ PHP_METHOD(amqp_exchange_class, declareExchange)
 		return;
 	}
 
-	arguments = convert_zval_to_arguments(exchange->arguments);
+	arguments = convert_zval_to_amqp_table(exchange->arguments TSRMLS_CC);
 	
+#if AMQP_VERSION_MAJOR * 100 + AMQP_VERSION_MINOR * 10 + AMQP_VERSION_PATCH > 52
 	amqp_exchange_declare(
 		connection->connection_resource->connection_state,
 		channel->channel_id,
 		amqp_cstring_bytes(exchange->name),
 		amqp_cstring_bytes(exchange->type),
-		exchange->passive,
-		exchange->durable,
+		IS_PASSIVE(exchange->flags),
+		IS_DURABLE(exchange->flags),
+		IS_AUTODELETE(exchange->flags),
+		IS_INTERNAL(exchange->flags),
 		*arguments
 	);
+#else
+	amqp_exchange_declare(
+		connection->connection_resource->connection_state,
+		channel->channel_id,
+		amqp_cstring_bytes(exchange->name),
+		amqp_cstring_bytes(exchange->type),
+		IS_PASSIVE(exchange->flags),
+		IS_DURABLE(exchange->flags),
+		*arguments
+	);
+#endif
 
-	res = AMQP_RPC_REPLY_T_CAST amqp_get_rpc_reply(connection->connection_resource->connection_state);
+	amqp_rpc_reply_t res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
 
-	AMQP_EFREE_ARGUMENTS(arguments);
+	php_amqp_free_amqp_table(arguments);
 
 	/* handle any errors that occured outside of signals */
 	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
-		char str[256];
-		char ** pstr = (char **) &str;
-		amqp_error(res, pstr, connection, channel);
+		PHP_AMQP_INIT_ERROR_MESSAGE();
 
-		zend_throw_exception(amqp_exchange_exception_class_entry, *pstr, 0 TSRMLS_CC);
-		efree(*pstr);
+		php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+		php_amqp_zend_throw_exception(res, amqp_exchange_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+		PHP_AMQP_DESTROY_ERROR_MESSAGE();
 		return;
 	}
 
+	php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
 	RETURN_TRUE;
 }
 /* }}} */
@@ -546,12 +529,9 @@ PHP_METHOD(amqp_exchange_class, delete)
 	amqp_connection_object *connection;
 
 	char *name = 0;
-	int name_len = 0;
-	long flags = 0;
+	int   name_len = 0;
 
-	amqp_rpc_reply_t res;
-	amqp_exchange_delete_t s;
-	amqp_method_number_t method_ok = AMQP_EXCHANGE_DELETE_OK_METHOD;
+	long flags = 0;
 
 	if (zend_parse_method_parameters(ZEND_NUM_ARGS() TSRMLS_CC, getThis(), "O|sl", &id, amqp_exchange_class_entry, &name, &name_len, &flags) == FAILURE) {
 		return;
@@ -559,49 +539,40 @@ PHP_METHOD(amqp_exchange_class, delete)
 
 	exchange = (amqp_exchange_object *)zend_object_store_get_object(id TSRMLS_CC);
 
-	if (name_len) {
-		AMQP_SET_NAME(exchange, name);
-		s.ticket = 0;
-		s.exchange.len = name_len;
-		s.exchange.bytes = name;
-		s.if_unused = (AMQP_IFUNUSED & flags) ? 1 : 0;
-		s.nowait = 0;
-	} else {
-		s.ticket = 0;
-		s.exchange.len = exchange->name_len;
-		s.exchange.bytes = exchange->name;
-		s.if_unused = (AMQP_IFUNUSED & flags) ? 1 : 0;
-		s.nowait = 0;
-	}
-
 	channel = AMQP_GET_CHANNEL(exchange);
-	AMQP_VERIFY_CHANNEL(channel, "Could not declare exchange.");
+	AMQP_VERIFY_CHANNEL(channel, "Could not delete exchange.");
 
 	connection = AMQP_GET_CONNECTION(channel);
-	AMQP_VERIFY_CONNECTION(connection, "Could not declare exchange.");
+	AMQP_VERIFY_CONNECTION(connection, "Could not delete exchange.");
 
-	res = amqp_simple_rpc(
+ 	amqp_exchange_delete(
 		connection->connection_resource->connection_state,
 		channel->channel_id,
-		AMQP_EXCHANGE_DELETE_METHOD,
-		&method_ok, &s
+		amqp_cstring_bytes(name_len ? name : exchange->name),
+		(AMQP_IFUNUSED & flags) ? 1 : 0
 	);
 
-	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
-		char str[256];
-		char ** pstr = (char **) &str;
-		amqp_error(res, pstr, connection, channel);
+	amqp_rpc_reply_t res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
 
-		zend_throw_exception(amqp_exchange_exception_class_entry, *pstr, 0 TSRMLS_CC);
-		amqp_maybe_release_buffers(connection->connection_resource->connection_state);
+	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
+		PHP_AMQP_INIT_ERROR_MESSAGE();
+
+		php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+		php_amqp_zend_throw_exception(res, amqp_exchange_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+		PHP_AMQP_DESTROY_ERROR_MESSAGE();
 		return;
 	}
-	amqp_maybe_release_buffers(connection->connection_resource->connection_state);
+
+	php_amqp_maybe_release_buffers_on_channel(connection, channel);
 
 	RETURN_TRUE;
 }
 /* }}} */
 
+
 /* {{{ proto AMQPExchange::publish(string msg, [string key, [int flags, [array headers]]]);
 publish into Exchange
 */
@@ -609,16 +580,19 @@ PHP_METHOD(amqp_exchange_class, publish)
 {
 	zval *id;
 	zval *ini_arr = NULL;
-	zval** zdata;
+	zval** ppztmp;
+	zval *ini_arr_tmp = NULL;
+
+
 	amqp_exchange_object *exchange;
 	amqp_channel_object *channel;
 	amqp_connection_object *connection;
 
 	char *key_name = NULL;
-	int key_len = 0;
+	int   key_len  = 0;
 
 	char *msg;
-	int msg_len= 0;
+	int   msg_len= 0;
 
 	long flags = AMQP_NOPARAM;
 
@@ -629,239 +603,169 @@ PHP_METHOD(amqp_exchange_class, publish)
 
 	amqp_basic_properties_t props;
 
-	int r;
-	amqp_bytes_t abt0, abt1, abt2;
-
-
 	if (zend_parse_method_parameters(ZEND_NUM_ARGS() TSRMLS_CC, getThis(), "Os|sla", &id, amqp_exchange_class_entry, &msg, &msg_len, &key_name, &key_len, &flags, &ini_arr) == FAILURE) {
 		return;
 	}
 
 	exchange = (amqp_exchange_object *)zend_object_store_get_object(id TSRMLS_CC);
 
-	if (exchange->name_len < 0) {
-		zend_throw_exception(amqp_exchange_exception_class_entry, "Could not publish to exchange. Exchange name not set.", 0 TSRMLS_CC);
-		return;
+	/* By default (and for BC) content type is text/plain (may be skipped at all, then set props._flags to 0) */
+	props.content_type = amqp_cstring_bytes("text/plain");
+	props._flags       = AMQP_BASIC_CONTENT_TYPE_FLAG;
+
+	props.headers.entries = 0;
+
+	if (ini_arr) {
+		ALLOC_ZVAL(ini_arr_tmp);
+		MAKE_COPY_ZVAL(&ini_arr, ini_arr_tmp);
 	}
 
-	props._flags = AMQP_BASIC_CONTENT_TYPE_FLAG;
+	{
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF (ini_arr_tmp), "content_type", sizeof("content_type"), (void*)&ppztmp)) {
+		convert_to_string(*ppztmp);
 
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF (ini_arr), "content_type", sizeof("content_type"), (void*)&zdata)) {
-		convert_to_string(*zdata);
-	}
-	if (zdata && strlen(Z_STRVAL_PP(zdata)) > 0) {
-		props.content_type = amqp_cstring_bytes((char *)Z_STRVAL_PP(zdata));
-	} else {
-		props.content_type = amqp_cstring_bytes("text/plain");
-	}
-
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF (ini_arr), "content_encoding", sizeof("content_encoding"), (void*)&zdata)) {
-		convert_to_string(*zdata);
-	}
-	if (zdata && strlen(Z_STRVAL_PP(zdata)) > 0) {
-		props.content_encoding = amqp_cstring_bytes((char *)Z_STRVAL_PP(zdata));
-		props._flags += AMQP_BASIC_CONTENT_ENCODING_FLAG;
-	}
-
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF (ini_arr), "message_id", sizeof("message_id"), (void*)&zdata)) {
-		convert_to_string(*zdata);
-	}
-	if (zdata && strlen(Z_STRVAL_PP(zdata)) > 0) {
-		props.message_id = amqp_cstring_bytes((char *)Z_STRVAL_PP(zdata));
-		props._flags += AMQP_BASIC_MESSAGE_ID_FLAG;
-	}
-
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF (ini_arr), "user_id", sizeof("user_id"), (void*)&zdata)) {
-		convert_to_string(*zdata);
-	}
-	if (zdata && strlen(Z_STRVAL_PP(zdata)) > 0) {
-		props.user_id = amqp_cstring_bytes((char *)Z_STRVAL_PP(zdata));
-		props._flags += AMQP_BASIC_USER_ID_FLAG;
-	}
-
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF (ini_arr), "app_id", sizeof("app_id"), (void*)&zdata)) {
-		convert_to_string(*zdata);
-	}
-	if (zdata && strlen(Z_STRVAL_PP(zdata)) > 0) {
-		props.app_id = amqp_cstring_bytes((char *)Z_STRVAL_PP(zdata));
-		props._flags += AMQP_BASIC_APP_ID_FLAG;
-	}
-
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF (ini_arr), "delivery_mode", sizeof("delivery_mode"), (void*)&zdata)) {
-		convert_to_long(*zdata);
-	}
-	if (zdata) {
-		props.delivery_mode = (uint8_t)Z_LVAL_PP(zdata);
-		props._flags += AMQP_BASIC_DELIVERY_MODE_FLAG;
-	}
-
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF (ini_arr), "priority", sizeof("priority"), (void*)&zdata)) {
-		convert_to_long(*zdata);
-	}
-	if (zdata) {
-		props.priority = (uint8_t)Z_LVAL_PP(zdata);
-		props._flags += AMQP_BASIC_PRIORITY_FLAG;
-	}
-
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF (ini_arr), "timestamp", sizeof("timestamp"), (void*)&zdata)) {
-		convert_to_long(*zdata);
-	}
-	if (zdata) {
-		props.timestamp = (uint64_t)Z_LVAL_PP(zdata);
-		props._flags += AMQP_BASIC_TIMESTAMP_FLAG;
-	}
-
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF (ini_arr), "expiration", sizeof("expiration"), (void*)&zdata)) {
-		convert_to_string(*zdata);
-	}
-	if (zdata && strlen(Z_STRVAL_PP(zdata)) > 0) {
-		props.expiration =	amqp_cstring_bytes((char *)Z_STRVAL_PP(zdata));
-		props._flags += AMQP_BASIC_EXPIRATION_FLAG;
-	}
-
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF (ini_arr), "type", sizeof("type"), (void*)&zdata)) {
-		convert_to_string(*zdata);
-	}
-	if (zdata && strlen(Z_STRVAL_PP(zdata)) > 0) {
-		props.type =  amqp_cstring_bytes((char *)Z_STRVAL_PP(zdata));
-		props._flags += AMQP_BASIC_TYPE_FLAG;
-	}
-
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF (ini_arr), "reply_to", sizeof("reply_to"), (void*)&zdata)) {
-		convert_to_string(*zdata);
-	}
-	if (zdata && strlen(Z_STRVAL_PP(zdata)) > 0) {
-		props.reply_to = amqp_cstring_bytes((char *)Z_STRVAL_PP(zdata));
-		props._flags += AMQP_BASIC_REPLY_TO_FLAG;
-	}
-
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF (ini_arr), "correlation_id", sizeof("correlation_id"), (void*)&zdata)) {
-		convert_to_string(*zdata);
-	}
-	if (zdata && strlen(Z_STRVAL_PP(zdata)) > 0) {
-		props.correlation_id = amqp_cstring_bytes((char *)Z_STRVAL_PP(zdata));
-		props._flags += AMQP_BASIC_CORRELATION_ID_FLAG;
-	}
-
-	zdata = NULL;
-	if (ini_arr && SUCCESS == zend_hash_find(HASH_OF(ini_arr), "headers", sizeof("headers"), (void*)&zdata)) {
-		HashTable *headers;
-		HashPosition pos;
-
-		convert_to_array(*zdata);
-		headers = HASH_OF(*zdata);
-		zend_hash_internal_pointer_reset_ex(headers, &pos);
-
-		props._flags += AMQP_BASIC_HEADERS_FLAG;
-		props.headers.entries = emalloc(sizeof(struct amqp_table_entry_t_) * zend_hash_num_elements(headers));
-		props.headers.num_entries = 0;
-
-		while (zend_hash_get_current_data_ex(headers, (void **)&zdata, &pos) == SUCCESS) {
-			char *string_key;
-			uint string_key_len;
-			int	type;
-			ulong  num_key;
-
-			type = zend_hash_get_current_key_ex(headers, &string_key, &string_key_len, &num_key, 0, &pos);
-
-			props.headers.entries[props.headers.num_entries].key.bytes = string_key;
-			props.headers.entries[props.headers.num_entries].key.len = string_key_len - 1;
-
-			if (Z_TYPE_P(*zdata) == IS_STRING) {
-				convert_to_string(*zdata);
-				props.headers.entries[props.headers.num_entries].value.kind = AMQP_FIELD_KIND_UTF8;
-				props.headers.entries[props.headers.num_entries].value.value.bytes.bytes = Z_STRVAL_P(*zdata);
-				props.headers.entries[props.headers.num_entries].value.value.bytes.len = Z_STRLEN_P(*zdata);
-				props.headers.num_entries++;
-			} else if (Z_TYPE_P(*zdata) == IS_LONG) {
-				convert_to_long(*zdata);
-				props.headers.entries[props.headers.num_entries].value.kind = AMQP_FIELD_KIND_I32;
-				props.headers.entries[props.headers.num_entries].value.value.i32 = Z_LVAL_P(*zdata);
-				props.headers.num_entries++;
-			} else if (Z_TYPE_P(*zdata) == IS_DOUBLE) {
-				convert_to_double(*zdata);
-				props.headers.entries[props.headers.num_entries].value.kind = AMQP_FIELD_KIND_F32;
-				props.headers.entries[props.headers.num_entries].value.value.f32 = (float)Z_DVAL_P(*zdata);
-				props.headers.num_entries++;
-			} else if (Z_TYPE_PP(zdata) == IS_ARRAY) {
-				zval **arr_data;
-				amqp_array_t array;
-				HashPosition arr_pos;
-				array.entries = emalloc(sizeof(struct amqp_field_value_t_) * zend_hash_num_elements(Z_ARRVAL_PP(zdata)));
-				array.num_entries = 0;
-				for(
-					zend_hash_internal_pointer_reset_ex(Z_ARRVAL_PP(zdata), &arr_pos);
-					zend_hash_get_current_data_ex(Z_ARRVAL_PP(zdata), (void**) &arr_data, &arr_pos) == SUCCESS;
-					zend_hash_move_forward_ex(Z_ARRVAL_PP(zdata), &arr_pos)
-				) {
-					if (Z_TYPE_PP(arr_data) == IS_STRING) {
-						array.entries[array.num_entries].kind = AMQP_FIELD_KIND_UTF8;
-						array.entries[array.num_entries].value.bytes.bytes = Z_STRVAL_PP(arr_data);
-						array.entries[array.num_entries].value.bytes.len = Z_STRLEN_PP(arr_data);
-						array.num_entries ++;
-					} else {
-						php_error_docref(NULL TSRMLS_CC, E_WARNING, "Ignoring non-string array member type %d for field '%s'", Z_TYPE_PP(arr_data), string_key);
-					}
-				}
-
-				props.headers.entries[props.headers.num_entries].value.kind = AMQP_FIELD_KIND_ARRAY;
-				props.headers.entries[props.headers.num_entries].value.value.array = array;
-				props.headers.num_entries++;
-			}
-
-			zend_hash_move_forward_ex(headers, &pos);
+		if (Z_STRLEN_PP(ppztmp) > 0) {
+			props.content_type = amqp_cstring_bytes((char *)Z_STRVAL_PP(ppztmp));
+			props._flags |= AMQP_BASIC_CONTENT_TYPE_FLAG;
 		}
-	} else {
-		props.headers.entries = 0;
+	}
+
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF (ini_arr_tmp), "content_encoding", sizeof("content_encoding"), (void*)&ppztmp)) {
+		convert_to_string(*ppztmp);
+
+		if (Z_STRLEN_PP(ppztmp) > 0) {
+			props.content_encoding = amqp_cstring_bytes((char *)Z_STRVAL_PP(ppztmp));
+			props._flags |= AMQP_BASIC_CONTENT_ENCODING_FLAG;
+		}
+	}
+
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF (ini_arr_tmp), "message_id", sizeof("message_id"), (void*)&ppztmp)) {
+		convert_to_string(*ppztmp);
+
+		if (Z_STRLEN_PP(ppztmp) > 0) {
+			props.message_id = amqp_cstring_bytes((char *)Z_STRVAL_PP(ppztmp));
+			props._flags |= AMQP_BASIC_MESSAGE_ID_FLAG;
+		}
+	}
+
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF (ini_arr_tmp), "user_id", sizeof("user_id"), (void*)&ppztmp)) {
+		convert_to_string(*ppztmp);
+
+		if (Z_STRLEN_PP(ppztmp) > 0) {
+			props.user_id = amqp_cstring_bytes((char *)Z_STRVAL_PP(ppztmp));
+			props._flags |= AMQP_BASIC_USER_ID_FLAG;
+		}
+	}
+
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF (ini_arr_tmp), "app_id", sizeof("app_id"), (void*)&ppztmp)) {
+		convert_to_string(*ppztmp);
+
+		if (Z_STRLEN_PP(ppztmp) > 0) {
+			props.app_id = amqp_cstring_bytes((char *)Z_STRVAL_PP(ppztmp));
+			props._flags |= AMQP_BASIC_APP_ID_FLAG;
+		}
+	}
+
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF (ini_arr_tmp), "delivery_mode", sizeof("delivery_mode"), (void*)&ppztmp)) {
+		convert_to_long(*ppztmp);
+
+		props.delivery_mode = (uint8_t)Z_LVAL_PP(ppztmp);
+		props._flags |= AMQP_BASIC_DELIVERY_MODE_FLAG;
+	}
+
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF (ini_arr_tmp), "priority", sizeof("priority"), (void*)&ppztmp)) {
+		convert_to_long(*ppztmp);
+
+		props.priority = (uint8_t)Z_LVAL_PP(ppztmp);
+		props._flags |= AMQP_BASIC_PRIORITY_FLAG;
+	}
+
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF (ini_arr_tmp), "timestamp", sizeof("timestamp"), (void*)&ppztmp)) {
+		convert_to_long(*ppztmp);
+
+		props.timestamp = (uint64_t)Z_LVAL_PP(ppztmp);
+		props._flags |= AMQP_BASIC_TIMESTAMP_FLAG;
+	}
+
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF (ini_arr_tmp), "expiration", sizeof("expiration"), (void*)&ppztmp)) {
+		convert_to_string(*ppztmp);
+
+		if (Z_STRLEN_PP(ppztmp) > 0) {
+			props.expiration = amqp_cstring_bytes((char *)Z_STRVAL_PP(ppztmp));
+			props._flags |= AMQP_BASIC_EXPIRATION_FLAG;
+		}
+	}
+
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF (ini_arr_tmp), "type", sizeof("type"), (void*)&ppztmp)) {
+		convert_to_string(*ppztmp);
+
+		if (Z_STRLEN_PP(ppztmp) > 0) {
+			props.type = amqp_cstring_bytes((char *)Z_STRVAL_PP(ppztmp));
+			props._flags |= AMQP_BASIC_TYPE_FLAG;
+		}
+	}
+
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF (ini_arr_tmp), "reply_to", sizeof("reply_to"), (void*)&ppztmp)) {
+		convert_to_string(*ppztmp);
+
+		if (Z_STRLEN_PP(ppztmp) > 0) {
+			props.reply_to = amqp_cstring_bytes((char *)Z_STRVAL_PP(ppztmp));
+			props._flags |= AMQP_BASIC_REPLY_TO_FLAG;
+		}
+	}
+
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF (ini_arr_tmp), "correlation_id", sizeof("correlation_id"), (void*)&ppztmp)) {
+		convert_to_string(*ppztmp);
+
+		if (Z_STRLEN_PP(ppztmp) > 0) {
+			props.correlation_id = amqp_cstring_bytes((char *)Z_STRVAL_PP(ppztmp));
+			props._flags |= AMQP_BASIC_CORRELATION_ID_FLAG;
+		}
+	}
+
+	}
+
+	amqp_table_t *headers = NULL;
+
+	if (ini_arr_tmp && SUCCESS == zend_hash_find(HASH_OF(ini_arr_tmp), "headers", sizeof("headers"), (void*)&ppztmp)) {
+		convert_to_array(*ppztmp);
+
+		headers = convert_zval_to_amqp_table(*ppztmp TSRMLS_CC);
+
+		props._flags |= AMQP_BASIC_HEADERS_FLAG;
+		props.headers = *headers;
 	}
 
 	channel = AMQP_GET_CHANNEL(exchange);
-	AMQP_VERIFY_CHANNEL(channel, "Could not publish to exchange exchange.");
+	AMQP_VERIFY_CHANNEL(channel, "Could not publish to exchange.");
 
 	connection = AMQP_GET_CONNECTION(channel);
-	AMQP_VERIFY_CONNECTION(connection, "Could not publish to exchange exchange.");
+	AMQP_VERIFY_CONNECTION(connection, "Could not publish to exchange.");
 
 #ifndef PHP_WIN32
 	/* Start ignoring SIGPIPE */
 	old_handler = signal(SIGPIPE, SIG_IGN);
 #endif
 
-	abt0.len = exchange->name_len;
-	abt0.bytes = exchange->name;
-	abt1.len = key_len;
-	abt1.bytes = key_name;
-	abt2.len = msg_len;
-	abt2.bytes = msg;
-
-	r = amqp_basic_publish(
+	/* NOTE: basic.publish is asynchronous and thus will not indicate failure if something goes wrong on the broker */
+	int status = amqp_basic_publish(
 		connection->connection_resource->connection_state,
 		channel->channel_id,
-		abt0,
-		abt1,
+		(exchange->name_len > 0 ? amqp_cstring_bytes(exchange->name) : amqp_empty_bytes),	/* exchange */
+		(key_len > 0 ? amqp_cstring_bytes(key_name) : amqp_empty_bytes), /* routing key */
 		(AMQP_MANDATORY & flags) ? 1 : 0, /* mandatory */
 		(AMQP_IMMEDIATE & flags) ? 1 : 0, /* immediate */
 		&props,
-		abt2
+		php_amqp_long_string(msg, msg_len) /* message body */
 	);
 
-	if (props.headers.entries) {
-		int i;
-		for (i=0; i < props.headers.num_entries; ++i) {
-			free_field_value(props.headers.entries[i].value);
-		}
-		efree(props.headers.entries);
+	if (headers) {
+		php_amqp_free_amqp_table(headers);
+	}
+
+	if (ini_arr_tmp) {
+		zval_dtor(ini_arr_tmp);
+		FREE_ZVAL(ini_arr_tmp);
 	}
 
 #ifndef PHP_WIN32
@@ -869,12 +773,20 @@ PHP_METHOD(amqp_exchange_class, publish)
 	signal(SIGPIPE, old_handler);
 #endif
 
-	/* handle any errors that occured outside of signals */
-	if (r < 0) {
-		char str[256];
-		char ** pstr = (char **) &str;
-        spprintf(pstr, 0, "Socket error: %s", amqp_error_string2(-r));
-        zend_throw_exception(amqp_exchange_exception_class_entry, *pstr, 0 TSRMLS_CC);
+	if (status != AMQP_STATUS_OK) {
+		/* Emulate library error */
+		amqp_rpc_reply_t res;
+		res.reply_type 	  = AMQP_RESPONSE_LIBRARY_EXCEPTION;
+		res.library_error = status;
+
+		PHP_AMQP_INIT_ERROR_MESSAGE();
+
+		php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+		php_amqp_zend_throw_exception(res, amqp_queue_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+		PHP_AMQP_DESTROY_ERROR_MESSAGE();
 		return;
 	}
 
@@ -883,27 +795,27 @@ PHP_METHOD(amqp_exchange_class, publish)
 /* }}} */
 
 
-/* {{{ proto int exchange::bind(string srcExchangeName, string routingKey[, int flags]);
+/* {{{ proto int exchange::bind(string srcExchangeName[, string routingKey, array arguments]);
 bind exchange to exchange by routing key
 */
 PHP_METHOD(amqp_exchange_class, bind)
 {
-	zval *id;
+	zval *id, *zvalArguments = NULL;
 	amqp_exchange_object *exchange;
 	amqp_channel_object *channel;
 	amqp_connection_object *connection;
 
 	char *src_name;
-	int src_name_len;
+	int   src_name_len = 0;
+
 	char *keyname;
-	int keyname_len;
+	int   keyname_len = 0;
+
 	int flags;
 
-	amqp_rpc_reply_t res;
-	amqp_exchange_bind_t s;
-	amqp_method_number_t method_ok = AMQP_EXCHANGE_BIND_OK_METHOD;
+	amqp_table_t *arguments;
 
-	if (zend_parse_method_parameters(ZEND_NUM_ARGS() TSRMLS_CC, getThis(), "Oss|l", &id, amqp_exchange_class_entry, &src_name, &src_name_len, &keyname, &keyname_len, &flags) == FAILURE) {
+	if (zend_parse_method_parameters(ZEND_NUM_ARGS() TSRMLS_CC, getThis(), "Os|sa", &id, amqp_exchange_class_entry, &src_name, &src_name_len, &keyname, &keyname_len, &zvalArguments) == FAILURE) {
 		return;
 	}
 
@@ -915,40 +827,107 @@ PHP_METHOD(amqp_exchange_class, bind)
 	connection = AMQP_GET_CONNECTION(channel);
 	AMQP_VERIFY_CONNECTION(connection, "Could not bind to exchanges.");
 
-	if (!keyname_len) {
-		zend_throw_exception(amqp_exchange_exception_class_entry, "Could not bind exchange. No routing key given.", 0 TSRMLS_CC);
-		return;
+	if (zvalArguments) {
+		arguments = convert_zval_to_amqp_table(zvalArguments TSRMLS_CC);
 	}
 
-	s.ticket				= 0;
-	s.destination.len		= exchange->name_len;
-	s.destination.bytes		= exchange->name;
-	s.source.len			= src_name_len;
-	s.source.bytes			= src_name;
-	s.routing_key.len		= keyname_len;
-	s.routing_key.bytes		= keyname;
-	s.nowait				= IS_NOWAIT(flags);
-	s.arguments.num_entries = 0;
-	s.arguments.entries		= NULL;
-
-	res = amqp_simple_rpc(
+	amqp_exchange_bind(
 		connection->connection_resource->connection_state,
 		channel->channel_id,
-		AMQP_EXCHANGE_BIND_METHOD,
-		&method_ok,
-		&s
+		amqp_cstring_bytes(exchange->name),
+		(src_name_len > 0 ? amqp_cstring_bytes(src_name) : amqp_empty_bytes),
+		(keyname_len  > 0 ? amqp_cstring_bytes(keyname)  : amqp_empty_bytes),
+		(zvalArguments ? *arguments : amqp_empty_table)
 	);
 
-	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
-		char str[256];
-		char ** pstr = (char **) &str;
-		amqp_error(res, pstr, connection, channel);
+	if (zvalArguments) {
+		php_amqp_free_amqp_table(arguments);
+	}
 
-		zend_throw_exception(amqp_exchange_exception_class_entry, *pstr, 0 TSRMLS_CC);
-		amqp_maybe_release_buffers(connection->connection_resource->connection_state);
+	amqp_rpc_reply_t res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
+
+	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
+		PHP_AMQP_INIT_ERROR_MESSAGE();
+
+		php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+		php_amqp_zend_throw_exception(res, amqp_exchange_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+		PHP_AMQP_DESTROY_ERROR_MESSAGE();
 		return;
 	}
-	amqp_maybe_release_buffers(connection->connection_resource->connection_state);
+
+	php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+	RETURN_TRUE;
+}
+/* }}} */
+
+/* {{{ proto int exchange::unbind(string srcExchangeName[, string routingKey, array arguments]);
+remove exchange to exchange binding by routing key
+*/
+PHP_METHOD(amqp_exchange_class, unbind)
+{
+	zval *id, *zvalArguments = NULL;
+	amqp_exchange_object *exchange;
+	amqp_channel_object *channel;
+	amqp_connection_object *connection;
+
+	char *src_name;
+	int   src_name_len = 0;
+
+	char *keyname;
+	int   keyname_len = 0;
+
+	int flags;
+
+	amqp_table_t *arguments;
+
+	if (zend_parse_method_parameters(ZEND_NUM_ARGS() TSRMLS_CC, getThis(), "Os|sa", &id, amqp_exchange_class_entry, &src_name, &src_name_len, &keyname, &keyname_len, &zvalArguments) == FAILURE) {
+		return;
+	}
+
+	exchange = (amqp_exchange_object *)zend_object_store_get_object(id TSRMLS_CC);
+
+	channel = AMQP_GET_CHANNEL(exchange);
+	AMQP_VERIFY_CHANNEL(channel, "Could not unbind from exchange.");
+
+	connection = AMQP_GET_CONNECTION(channel);
+	AMQP_VERIFY_CONNECTION(connection, "Could not unbind from exchanges.");
+
+	if (zvalArguments) {
+		arguments = convert_zval_to_amqp_table(zvalArguments TSRMLS_CC);
+	}
+
+	amqp_exchange_unbind(
+		connection->connection_resource->connection_state,
+		channel->channel_id,
+		amqp_cstring_bytes(exchange->name),
+		(src_name_len > 0 ? amqp_cstring_bytes(src_name) : amqp_empty_bytes),
+		(keyname_len  > 0 ? amqp_cstring_bytes(keyname)  : amqp_empty_bytes),
+		(zvalArguments ? *arguments : amqp_empty_table)
+	);
+
+	if (zvalArguments) {
+		php_amqp_free_amqp_table(arguments);
+	}
+
+	amqp_rpc_reply_t res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
+
+	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
+		PHP_AMQP_INIT_ERROR_MESSAGE();
+
+		php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+		php_amqp_zend_throw_exception(res, amqp_exchange_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+		PHP_AMQP_DESTROY_ERROR_MESSAGE();
+		return;
+	}
+
+	php_amqp_maybe_release_buffers_on_channel(connection, channel);
 
 	RETURN_TRUE;
 }
@@ -967,7 +946,7 @@ PHP_METHOD(amqp_exchange_class, getChannel)
 
 	exchange = (amqp_exchange_object *)zend_object_store_get_object(id TSRMLS_CC);
 
-    RETURN_ZVAL(exchange->channel, 1, 0);
+	RETURN_ZVAL(exchange->channel, 1, 0);
 }
 /* }}} */
 
@@ -986,7 +965,7 @@ PHP_METHOD(amqp_exchange_class, getConnection)
 	exchange = (amqp_exchange_object *)zend_object_store_get_object(id TSRMLS_CC);
 	channel = AMQP_GET_CHANNEL(exchange);
 
-    RETURN_ZVAL(channel->connection, 1, 0);
+	RETURN_ZVAL(channel->connection, 1, 0);
 }
 /* }}} */
 
